/************************************************************************
 *                                                                      *
 * File:        my_rad.c		                                *
 *                                                                      *
 * Author(s):   T. Frivold, D.Beyer.  Rooftop Communications.           *
 *                                                                      *
 * Description: Skeletal code for Radio Device                          *
 *                                                                      *
 ************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <my_rad.h>
#include <my_radpr.h>	/* Private, hardware specific functions/values */

/**********************************************************************
 *
 * static variables and function protototypes for MyRadio
 *
 **********************************************************************/

/*
 * MyRadio Entry Points
 */
static uint32 _MyRadioOpen      ( void * devData, bytep name, bytep mode );
static uint32 _MyRadioClose     ( void * devData );
static uint32 _MyRadioSigEnable ( void * devData, uint32 num, uint32 flag );
static uint32 _MyRadioCmd       ( void * devData, uint32 num, uint32 qual,
				  void * data,    uint32 dataLen );
static uint32 _MyRadioVar       ( void * devData, uint32 num, uint32 qual,
				  void * data,    uint32 dataLen );

/*
 * MyRadio's interrupt handler 
 */
static void    _HandleInterrupt  ( RadioDevData * dev, int32 intNum, 
				   void * data,        int32 len );

/**********************************************************************
 *
 * MyRadioDevInit()
 *
 **********************************************************************/

uint32
MyRadioDevInit ( uint32        version,
		 void        * protoDev,
		 void        * initData,
		 DevSigFp      sig,     
		 DevCntrlBlk * control )
{
    RadioDevData * radioDevData;
    uint32	   i;

    /*
     * Check for proper version
     */
    if (version != RADIO_API_VERSION) 
    {
	/* First entry of structure remains valid across versions. 	*/
	/* So, it's ok to initialize it before returning		*/
	control->version  = RADIO_API_VERSION;

	/* If the protocol sware can suppport my version, it will	*/
	/* call again passing the version number that I'm built for.	*/

	return RadioRetInvVersion; 
    }

    /*
     * Allocate a device structure
     */

    radioDevData = (RadioDevData *) malloc( sizeof(RadioDevData) );
    if ( !radioDevData ) return RadioRetMemOut;

    /*
     * Initialize radio data structure
     */

    radioDevData->protoDev  = protoDev;
    radioDevData->sigReport = sig;
    radioDevData->txPower   = MyRadioTxPwr30_dBm;

    for (i = 0; i < 6; i++) radioDevData->freq[i] = 0;

    /*
     * Fill in the control block for the user
     */

    control->version     = RADIO_API_VERSION;
    control->devHandle   = (void *) radioDevData;
    control->maxSigNum   = RADIO_SIG_NEXT - 1;

    /* main control entry points */
    control->cmd         = _MyRadioCmd;
    control->var         = _MyRadioVar;

    /* other standard device entry points */
    control->open        = _MyRadioOpen;
    control->close       = _MyRadioClose;
    control->sigEnable   = _MyRadioSigEnable;
    control->idle        = (void *) 0;   /* No idle function  */
    control->read        = (void *) 0;   /* No read function  */
    control->write       = (void *) 0;   /* No write function */
    control->flush       = (void *) 0;   /* No flush function */

    return RadioRetOk;
}

/**********************************************************************
 *
 * MyRadioOpen()
 *
 **********************************************************************/

static uint32
_MyRadioOpen ( void * radioDev, bytep name, bytep mode )
{
    RadioDevData * radioDevData = (RadioDevData *) radioDev;

    strncpy ( radioDevData->name, name, 128 );
    
    _RegisterInteruptHandler ( _HandleInterrupt, radioDev /* other args.. */ );

    /* ... do other stuff to activate the radio device */

    return RadioRetOk;
}

/**********************************************************************
 *
 * MyRadioClose()
 *
 **********************************************************************/

static uint32
_MyRadioClose ( void * radioDev )
{
    /* ... do stuff to deactivate the radio device */

    return RadioRetOk;
}

/**********************************************************************
 *
 * MyRadioSigEnable()
 *
 **********************************************************************/

static uint32
_MyRadioSigEnable ( void * radioDev, uint32 sigNum, uint32 flag )
{
    switch ( sigNum )
    {
    case RadioSigAll:
	if (flag == RadioSigEnable)
	{
	    _InterruptEnable  (interruptRcvCarrierOn);
	    _InterruptEnable  (interruptRcvCarrierOff);
	    /* ... */
	} 
	else 
	{
	    _InterruptDisable  (interruptRcvCarrierOn);
	    _InterruptDisable  (interruptRcvCarrierOff);
	    /* ... */
	}
	break;

    case RadioSigCarrierActive:

        if (flag == RadioSigEnable ) _InterruptEnable  (interruptRcvCarrierOn);
        else                         _InterruptDisable (interruptRcvCarrierOn);
        break;

    case RadioSigCarrierInactive:

        if (flag == RadioSigEnable) _InterruptEnable  (interruptRcvCarrierOff);
        else                        _InterruptDisable (interruptRcvCarrierOff);
        break;

    /* ... do for other signal that can be enabled/disabled */

    default:
	return RadioRetInvSig;

    }

    return RadioRetOk;
}

/**********************************************************************
 *
 * MyRadioCmd()
 *
 **********************************************************************/
static uint32
_MyRadioCmd ( void * radioDev, uint32 cmdNum, uint32 cmdQual,
	      void * data,     uint32 dataLen )
{
    uint32 retCode = RadioRetOk;
    MyRadioPktInfo * pktData;

    switch (cmdNum)
    {

    case RadioCmdReset:

	/* reset the radio... */

	break;

    case RadioCmdXmtPkt:

	/* Return if MyRadioPktInfo size structure wasn't passed.	*/ 
	/* Or, if only RadioPktInfo size, then could proceed and 	*/
	/* ignore the MyRadioPktInfo extension fields 			*/

	if (dataLen != sizeof(MyRadioPktInfo)) return RadioRetInvParam;

	/* 
	 * Here we assume an example where the device can handle
	 * the transmission of multple-pkt bursts, and the pkt transmission 
	 * parameters can be changed on an individual pkt-by-pkt basis 
	 * (rather than for an entire burst of packets).  
	 * So, a the transmission instructions passed down in the 
	 * MyRadioPktInfo structure must be copied into a local copy, which
	 * will be tied to the packet for reference until it is ready to 
	 * be transmitted.
	 */

        /* get a pkt info struct from a local pool */
        pktData = _PktInfoPoolGet (); 

	/* copy user's structure into local pool copy */
        memcpy ( pktData, data, sizeof(MyRadioPktInfo) );

	/*
	 * Start transmitting this packet
	 * if not already transmitting a packet, else enqueue for 
	 * later transmission.
	 */

	if ( _TransmitterOff() )  
	{
	    _TurnOnTransmitter();
	    _XmtPkt ( pktData );
	}
	else
	{
	    _XmtQEnqueue ( pktData ); 
	}

	break;

    case RadioCmdRcvPkt:

        /* get a pkt info struct from the pool */
        pktData = _PktInfoPoolGet (); 

	/*
	 * make sure that dataLen == sizeof(MyRadioPktInfo) 
	 * (or, if less, then ignore MyRadio extension fields ...)
	 */

	/* copy user's structure into ours (not the actual buffer!) */
        memcpy ( pktData, data, sizeof(MyRadioPktInfo) );

	/*
	 * Enqueue on the radio's receive pkt queue.  This pointer
	 * will later be handed back through the
	 * _HandleInterrupt() callback when a pkt has been received.
	 */

	_RcvQEnqueue ( pktData );

	break;

    /* ... do for other commands */

    default:
	return RadioRetInvCmd;

    }

    return retCode;
}

/**********************************************************************
 *
 * MyRadioVar()
 *
 **********************************************************************/
static uint32
_MyRadioVar ( void * radioDev, uint32 varNum, uint32 varQual,
	      void * data,     uint32 dataLen )
{
    RadioDevData * radioDevData = (RadioDevData *) radioDev;
    uint32         chNum;
    int32	   value;


    switch ( varNum )
    {

    case RadioVarXmtBurstCnt:

	if (dataLen != DEV_ARG_LEN_INT32P) return RadioRetInvParam;
        
        _DisableInterrupts();

	value = radioDevData->burstCnt;

        if (varQual & RADIO_QUAL_INC)
	{
	    /* increment operation */
	    radioDevData->burstCnt += *((int32 *)data);

	    /* don't allow burst count to become negative */
	    if (radioDevData->burstCnt < 0) radioDevData->burstCnt = 0;
	}

	/* Notice that both inc and get can be performed in the  	*/
        /* same operation 						*/

        if (varQual & RADIO_QUAL_GET)
	{
	    /* return original value */
	    *((int32 *)data) = value;
	}

	_EnableInterrupts();

	break;

    case RadioVarFreq:

	if (dataLen != DEV_ARG_LEN_INT32P) return RadioRetInvParam;

	chNum = RADIO_QUAL_CH_DECODE(varQual);

	value = radioDevData->freq[chNum];

	if (varQual & RADIO_QUAL_SET)
	{
	    /* set operation */
	    radioDevData->freq[chNum] = *((int32 *)data);
	}

	if (varQual & RADIO_QUAL_GET)
	{
	    /* return original value */
	    *((int32 *)data) = value;
	}

	break;

    case RadioVarVersion:

	if (varQual & RADIO_QUAL_GET)
	{
	    /* return current value */
	    byte str[100];

	    sprintf(str, "Skeletal radio device driver compiled %s", 
		       __DATE__);

	    strncpy ((bytep) data, str, (int32)dataLen-1);

	    ((bytep)data)[dataLen-1] = '\0';	/* just  to be sure */
	}

	break;


    /* ... do for other variables */
	
    default:
	return RadioRetInvVar;

    }

    return RadioRetOk;
}


/**********************************************************************
 *
 * MyRadio's Interrupt Handler
 *
 **********************************************************************/

static void
_HandleInterrupt ( RadioDevData * dev, int32 intNum, void * data, int32 len )
{
    MyRadioPktInfo * pktData;
    uint32           qual = 0;
    uint32	     retCode = RadioRetOk;

    switch (intNum)
    {

    case interruptXmtPkt:  /* (private interrupt names for my radio) */

        /* 
	 * Pkt transmission has completed (and pktInfo structure is returned)
	 */
        
        pktData = (MyRadioPktInfo *) data; 

	/* check for packet errors */
	if (pktData->radioPktInfo.errStatus != 0)
	{
	    retCode = RadioRetPktXmtError;
	}

        /*
	 * Call user's signal handler to return the pkt buffer to the user.
	 */

        (*(dev->sigReport)) ( dev->protoDev,
			      RadioSigXmtPkt,
			      qual,
			      (void *) pktData,
			      sizeof(MyRadioPktInfo),
			      retCode );

        _PktInfoPoolPut (pktData);  /* return to my pool */

	/* decrement burst count if > 0 */
	if (dev->burstCnt > 0) dev->burstCnt--;

	/*
	 * Check whether there are more packets to send.
	 */

	if ( dev->burstCnt == 0 && _XmtQEmpty() )
	{
	    /*
	     * If no packets remaining in burst and no packets
	     * left in the transmit queue, turn off transmitter.
	     */
               	    
	    _TurnOffTransmitter();

	    /*
  	     * If enabled, RadioSigXmtActive signal sent up 
             * later (after transmitter has powered-off)
	     */
	}
	else if ( !_XmtQEmpty() )
	{
	   /*
	    * Otherwise, if packets remain in the transmit
	    * queue, transmit the next packet in line.
	    */

	    _XmtPkt ( _XmtQDequeue() );
	}
	else
	{
	   /* Otherwise, the transmit queue is empty, but the
	    * burst count is not yet zero, so wait for more
	    * packets to be passed down and leave the transmitter 
	    * running (sending idle flags) until we have more data.
	    */
	}

        break;

    case interruptRcvPkt:

        /* pkt has been received */

	/* assumes pktInfo is passed in with all fields set */
        pktData = (MyRadioPktInfo *) data; 

	/* check for error */
	if (pktData->radioPktInfo.errStatus != 0)
	{
	    retCode = RadioRetPktRcvError;
	}
        
        /*
	 * Call user's signal handler to return the
	 * pkt buffer to the user
	 */

        (*(dev->sigReport)) ( dev->protoDev,
			      RadioSigRcvPkt,
			      qual,
			      (void *) pktData,
			      sizeof(MyRadioPktInfo),
			      retCode );

        _PktInfoPoolPut (data);  /* return to my pool */
	break;

    case interruptRcvCarrierOn:

        (*(dev->sigReport)) ( dev->protoDev,
			      RadioSigCarrierActive,
			      0, (void *) 0, 0, retCode );
        break;

    case interruptRcvCarrierOff: 

        (*(dev->sigReport)) ( dev->protoDev,
			      RadioSigCarrierInactive,
			      0, (void *) 0, 0, retCode );

        break;
    }
}


